@@ -1,4 +1,6 @@
-from molgraph import tensors
+from molgraph import _filter_warnings
+del _filter_warnings
+
 from molgraph import layers
 from molgraph import losses
 from molgraph import metrics
@@ -6,12 +8,6 @@ from molgraph import models
 from molgraph import chemistry
 
 from molgraph.tensors import GraphTensor
-from molgraph.tensors import GraphTensorSpec
-
-# filter out some warnings
-from molgraph import _filter_warnings
-
-del _filter_warnings
 
 from . import _version
 
