@@ -3,13 +3,6 @@ from molgraph.chemistry.molecular_encoders import MolecularGraphEncoder3D
 
 from molgraph.chemistry.conformer_generator import ConformerGenerator
 
-from molgraph.chemistry.encoders import AtomicFeaturizer
-from molgraph.chemistry.encoders import AtomicTokenizer
-from molgraph.chemistry.encoders import AtomFeaturizer
-from molgraph.chemistry.encoders import BondFeaturizer
-from molgraph.chemistry.encoders import AtomTokenizer
-from molgraph.chemistry.encoders import BondTokenizer
-
 from molgraph.chemistry.encoders import Tokenizer
 from molgraph.chemistry.encoders import Featurizer
 from molgraph.chemistry.features import Feature
