@@ -1,23 +1,29 @@
 import numpy as np
+
 from rdkit import Chem
+
 import tensorflow as tf
+
 from typing import Callable
 from typing import Union
 from typing import Optional
 from typing import List
-from typing import Any
 from typing import Callable
 from typing import Dict
 from typing import Tuple
 from typing import Sequence
-import multiprocessing
-from functools import partial
-import logging
-import collections
-import os
 
-from dataclasses import dataclass, field
-from abc import ABC, abstractmethod
+import multiprocessing
+
+from functools import partial
+
+import logging
+
+from dataclasses import dataclass
+from dataclasses import field
+
+from abc import ABC
+from abc import abstractmethod
 
 from molgraph.tensors.graph_tensor import GraphTensor
 from molgraph.chemistry.encoders import Featurizer
@@ -86,11 +92,7 @@ class BaseMolecularGraphEncoder(ABC):
                 graph_tensor_list = [
                     gt for gt in graph_tensor_list if gt is not None]
 
-                # The list of `GraphTensor`s is concatenated to generate a single
-                # `GraphTensor` (disjoint [molecular] graph). The `separate` method
-                # is called to make the nested structures of the `GraphTensor`
-                # ragged. This will allow for batching of the `GraphTensor`.
-                return tf.concat(graph_tensor_list, axis=0).separate()
+                return tf.concat(graph_tensor_list, axis=0)
 
             return self.call(inputs, **kwargs)
 
@@ -145,15 +147,14 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
     ... )
     >>> # Encode two molecules as a GraphTensor
     >>> graph_tensor = encoder(['CCC', 'CCO'])
-    >>> # Merge subgraphs into a single disjoint graph
-    >>> graph_tensor.merge()
+    >>> graph_tensor
     GraphTensor(
+      sizes=<tf.Tensor: shape=(2,), dtype=int32>,
       node_feature=<tf.Tensor: shape=(6, 119), dtype=float32>,
-      edge_feature=<tf.Tensor: shape=(8, 4), dtype=float32>,
-      positional_encoding=<tf.Tensor: shape=(6, 10), dtype=float32>,
-      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
       edge_src=<tf.Tensor: shape=(8,), dtype=int32>,
-      graph_indicator=<tf.Tensor: shape=(6,), dtype=int32>)
+      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
+      edge_feature=<tf.Tensor: shape=(8, 4), dtype=float32>,
+      node_position=<tf.Tensor: shape=(6, 10), dtype=float32>)
 
     Generate a molecular graph with tokenizers:
 
@@ -177,15 +178,14 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
     ... )
     >>> # Encode two molecules as a GraphTensor
     >>> graph_tensor = encoder(['CCC', 'CCO'])
-    >>> # Merge subgraphs into a single disjoint graph
-    >>> graph_tensor.merge()
+    >>> graph_tensor
     GraphTensor(
+      sizes=<tf.Tensor: shape=(2,), dtype=int32>,
       node_feature=<tf.Tensor: shape=(6,), dtype=string>,
-      edge_feature=<tf.Tensor: shape=(8,), dtype=string>,
-      positional_encoding=<tf.Tensor: shape=(6, 10), dtype=float32>,
-      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
       edge_src=<tf.Tensor: shape=(8,), dtype=int32>,
-      graph_indicator=<tf.Tensor: shape=(6,), dtype=int32>)
+      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
+      edge_feature=<tf.Tensor: shape=(8,), dtype=string>,
+      node_position=<tf.Tensor: shape=(6, 10), dtype=float32>)
 
     Obtain numerical encodings of atoms (``node_feature``) and bonds
     (``bond_feature``) with the EmbeddingLookup layer. This is only necessary
@@ -209,8 +209,6 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
     ... )
     >>> # Encode two molecules as a GraphTensor
     >>> graph_tensor = encoder(['CCC', 'CCO'])
-    >>> # Merge subgraphs into a single disjoint graph
-    >>> graph_tensor = graph_tensor.merge()
     >>> # Define embedding layers
     >>> node_embedding = molgraph.layers.EmbeddingLookup(
     ...    feature='node_feature', output_dim=16)
@@ -221,7 +219,6 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
     >>> edge_embedding.adapt(graph_tensor)
     >>> # Build model
     >>> model = tf.keras.Sequential([
-    ...    tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
     ...    node_embedding,
     ...    edge_embedding,
     ... ])
@@ -229,13 +226,12 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
     >>> graph_tensor = model(graph_tensor)
     >>> graph_tensor
     GraphTensor(
+      sizes=<tf.Tensor: shape=(2,), dtype=int32>,
       node_feature=<tf.Tensor: shape=(6, 16), dtype=float32>,
-      edge_feature=<tf.Tensor: shape=(8, 8), dtype=float32>,
-      positional_encoding=<tf.Tensor: shape=(6, 10), dtype=float32>,
-      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
       edge_src=<tf.Tensor: shape=(8,), dtype=int32>,
-      graph_indicator=<tf.Tensor: shape=(6,), dtype=int32>)
-
+      edge_dst=<tf.Tensor: shape=(8,), dtype=int32>,
+      edge_feature=<tf.Tensor: shape=(8, 8), dtype=float32>,
+      node_position=<tf.Tensor: shape=(6, 10), dtype=float32>)
     '''
 
     bond_encoder: Optional[Union[
@@ -277,12 +273,11 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
         # Obtain destination and source node (atom) indices of edges (bonds)
         sparse_adjacency = _compute_adjacency(
             molecule, self.self_loops, sparse=True, dtype=index_dtype)
-        data['edge_dst'], data['edge_src'] = sparse_adjacency
+        data['edge_src'], data['edge_dst'] = sparse_adjacency
 
         # Obtain node (atom) features
         atoms = _get_atoms(molecule)
         data['node_feature'] = self.atom_encoder(atoms)
-
         # Obtain edge (bond) features (if `bond_encoder` exist)
         if self.bond_encoder is not None:
             bonds = _get_bonds(molecule, *sparse_adjacency)
@@ -291,7 +286,7 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
 
         # Obtain positional encoding of nodes (atoms)
         if self.positional_encoding_dim:
-            data['positional_encoding'] = _compute_positional_encoding(
+            data['node_position'] = _compute_positional_encoding(
                 molecule=molecule,
                 dim=self.positional_encoding_dim,
                 dtype=getattr(self.atom_encoder, 'dtype', 'float32'))
@@ -300,7 +295,7 @@ class MolecularGraphEncoder(BaseMolecularGraphEncoder):
             for field, encoder in self.auxiliary_encoders.items():
                 data[field] = encoder(molecule)
 
-        return GraphTensor(data)
+        return GraphTensor(**data)
 
 
 @dataclass
@@ -345,8 +340,6 @@ class MolecularGraphEncoder3D(BaseMolecularGraphEncoder):
     ... )
     >>> # Encode two molecules as a GraphTensor
     >>> graph_tensor = encoder(['CCC', 'CCO'])
-    >>> # Merge subgraphs into a single disjoint graph.
-    >>> graph_tensor = graph_tensor.merge()
     >>> # The main difference between the 2d and 3d encoder is
     >>> # the edge_feature field. Here, in contains coulomb values,
     >>> # which mimics electrostatic interactions between nuclei
@@ -398,23 +391,23 @@ class MolecularGraphEncoder3D(BaseMolecularGraphEncoder):
 
         atoms = _get_atoms(molecule)
         data['node_feature'] = self.atom_encoder(atoms)
-        data['edge_dst'] = np.array(dg['edge_dst'], dtype=index_dtype)
         data['edge_src'] = np.array(dg['edge_src'], dtype=index_dtype)
-
+        data['edge_dst'] = np.array(dg['edge_dst'], dtype=index_dtype)
+        
         if not self.coulomb:
             edge_feature = np.expand_dims(dg['edge_length'], -1)
         else:
             nuclear_charge = np.array([
                 atom.GetAtomicNum() for atom in atoms], dtype=np.float32)
-            nuclear_charge_dst = np.take(nuclear_charge, dg['edge_dst'])
             nuclear_charge_src = np.take(nuclear_charge, dg['edge_src'])
+            nuclear_charge_dst = np.take(nuclear_charge, dg['edge_dst'])
             edge_feature = (
                 nuclear_charge_dst * nuclear_charge_src) / dg['edge_length']
             edge_feature = np.expand_dims(edge_feature, -1)
 
         data['edge_feature'] = np.array(edge_feature, dtype=np.float32)
 
-        return GraphTensor(data)
+        return GraphTensor(**data)
 
 
 def _get_atoms(molecule: Chem.Mol) -> List[Chem.Atom]:
@@ -423,8 +416,8 @@ def _get_atoms(molecule: Chem.Mol) -> List[Chem.Atom]:
 
 def _get_bonds(
     molecule: Chem.Mol,
+    edge_src: np.ndarray,
     edge_dst: np.ndarray,
-    edge_src: np.ndarray
 ) -> List[Chem.Bond]:
     '''Returns a list of bonds given an RDKit mol object. The order of the
     bonds in the list corresponds to the sparse adjacency matrix which is
@@ -432,7 +425,7 @@ def _get_bonds(
     '''
     return [
         molecule.GetBondBetweenAtoms(int(i), int(j))
-        for (i, j) in zip(edge_dst, edge_src)
+        for (i, j) in zip(edge_src, edge_dst)
     ]
 
 def _compute_positional_encoding(
@@ -502,9 +495,9 @@ def _compute_adjacency(
     if not sparse:
         return adjacency.astype(dtype)
 
-    bond_dst, bond_src = np.where(adjacency)
+    edge_src, edge_dst = np.where(adjacency)
 
-    return bond_dst.astype(dtype), bond_src.astype(dtype)
+    return edge_src.astype(dtype), edge_dst.astype(dtype)
 
 def _compute_laplacian(
     adjacency: np.ndarray,
@@ -521,16 +514,16 @@ def _compute_laplacian(
 
 def _compute_distance_between_atoms(
     molecule: Chem.Mol,
-    edge_dst: int,
     edge_src: int,
+    edge_dst: int,
     unit: str
 ) -> float:
     'Computes distance between two atoms in the molecule.'
-    bond_length = Chem.rdMolTransforms.GetBondLength(
-        molecule.GetConformer(), edge_dst, edge_src)
+    edge_length = Chem.rdMolTransforms.GetBondLength(
+        molecule.GetConformer(), edge_src, edge_dst)
     if unit.lower() == 'bohr':
-        return bond_length * 1.8897259885789
-    return bond_length
+        return edge_length * 1.8897259885789
+    return edge_length
 
 def _compute_distance_geometry(
     molecule: Chem.Mol,
@@ -548,18 +541,21 @@ def _compute_distance_geometry(
 
     if path is None:
         data = {
-            'edge_length': [], 'edge_dst': [], 'edge_src': [], 'edge_order': []}
+            'edge_length': [], 'edge_src': [], 'edge_dst': [], 'edge_order': []
+        }
         for atom in molecule.GetAtoms():
             path = _compute_distance_geometry(
-                molecule, radius, unit, atom, [atom.GetIdx()], data)
+                molecule, radius, unit, atom, [atom.GetIdx()], data
+            )
         return data
     elif radius and len(path) > (radius + 1):
         return path[:-1]
     elif len(path) > 1:
-        data['edge_dst'].append(path[0])
-        data['edge_src'].append(path[-1])
+        data['edge_src'].append(path[0])
+        data['edge_dst'].append(path[-1])
         data['edge_length'].append(
-            _compute_distance_between_atoms(molecule, path[0], path[-1], unit))
+            _compute_distance_between_atoms(molecule, path[0], path[-1], unit)
+        )
         data['edge_order'].append(-1 + len(path))
 
     for neighbor in atom.GetNeighbors():
