@@ -1,99 +1,69 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
 
+from keras import initializers
+from keras import regularizers
+from keras import constraints
+from keras import activations
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import softmax_edge_weights
-from molgraph.layers.ops import propagate_node_features
-from molgraph.layers.ops import reduce_features
+
+from molgraph.layers import gnn_layer
+from molgraph.layers import gnn_ops
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class GATv2Conv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class GATv2Conv(gnn_layer.GNNLayer):
 
     '''Multi-head graph attention (v2) layer (GATv2).
 
     The implementation is based on Brody et al. (2021) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...         'edge_feature': [
-    ...             [[1.0, 0.0], [0.0, 1.0]],
-    ...             [[0.0, 1.0], [0.0, 1.0], [1.0, 0.0],
-    ...              [0.0, 1.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GATv2Conv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GATv2Conv(16, activation='relu'),
-    ...     molgraph.layers.GATv2Conv(16, activation='relu')
+    ...     molgraph.layers.GATv2Conv(units=16),
+    ...     molgraph.layers.GATv2Conv(units=16),
+    ...     molgraph.layers.GATv2Conv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
+    
+    Including edge features:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...         'edge_feature': [
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0],
-    ...             [0.0, 1.0],
-    ...             [0.0, 1.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_feature=[[1., 0.], [0., 1.], [0., 1.], [0., 1.], 
+    ...                   [1., 0.], [0., 1.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GATv2Conv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GATv2Conv(16, activation='relu'),
-    ...     molgraph.layers.GATv2Conv(16, activation='relu')
+    ...     molgraph.layers.GATv2Conv(units=16, use_edge_features=True),
+    ...     molgraph.layers.GATv2Conv(units=16, use_edge_features=True),
+    ...     molgraph.layers.GATv2Conv(units=16, use_edge_features=True),
     ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> output = gnn_model(graph_tensor)
+    >>> output.node_feature.shape, output.edge_feature.shape
+    (TensorShape([5, 16]), TensorShape([8, 16]))
 
     Args:
         units (int, None):
             Number of output units.
-        use_edge_features (bool):
-            Whether or not to use edge features. Default to True.
         num_heads (int):
             Number of attention heads. Default to 8.
         merge_mode (str):
@@ -101,8 +71,9 @@ class GATv2Conv(BaseLayer):
             'mean' or None. If set to None, 'mean' is used. Default to 'concat'.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -132,7 +103,23 @@ class GATv2Conv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
 
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
+            *   `use_edge_features`: Whether or not to use edge features. 
+                Only relevant if edge features exist. If None, and edge 
+                features exist, it will be set to True. Default to None.
+            *   `update_edge_features` (bool): Specifies whether edge features 
+                should be updated along with node features, including the 
+                post-processing step. Only relevant if edge features exist. 
+                It is important that GNN layers which updates its edge features
+                for the next layer sets this to True. Default to False. 
     References:
         .. [#] https://arxiv.org/pdf/2105.14491.pdf
 
@@ -141,23 +128,18 @@ class GATv2Conv(BaseLayer):
     def __init__(
         self,
         units: Optional[int] = 128,
-        use_edge_features: bool = True,
         num_heads: int = 8,
         merge_mode: Optional[str] = 'concat',
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         attention_activation: Union[
             None, str, Callable[[tf.Tensor], tf.Tensor]] = 'leaky_relu',
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = 'relu',
         use_bias: bool = True,
-        kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.TruncatedNormal(stddev=0.005),
-        bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+        kernel_initializer: Union[str, initializers.Initializer, None] = None,
+        bias_initializer: Union[str, initializers.Initializer, None] = None,
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -166,11 +148,12 @@ class GATv2Conv(BaseLayer):
         **kwargs
     ):
         kwargs['update_edge_features'] = (
-            kwargs.get('update_edge_features', True) and use_edge_features
+            kwargs.get('update_edge_features', True) and 
+            kwargs.get('use_edge_features', True)
         )
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -184,17 +167,12 @@ class GATv2Conv(BaseLayer):
             bias_constraint=bias_constraint,
             **kwargs)
 
-        self.use_edge_features = use_edge_features
         self.num_heads = num_heads
         self.merge_mode = merge_mode
         self.apply_self_projection = self_projection
         self.attention_activation = activations.get(attention_activation)
 
-    def subclass_build(
-        self,
-        node_feature_shape: Optional[tf.TensorShape] = None,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
 
         if self.merge_mode == 'concat':
             if not self.units or (self.units % self.num_heads != 0):
@@ -203,23 +181,17 @@ class GATv2Conv(BaseLayer):
                     ' need `units` to be divisble by `num_heads`')
             self.units //= self.num_heads
 
-        self.use_edge_features = (
-            self.use_edge_features and edge_feature_shape is not None
-        )
         if self.use_edge_features:
-
-            self.edge_projection = self.get_einsum_dense(
-                'ij,jkh->ihk', (self.num_heads, self.units))
-
             if self.update_edge_features:
                 self.edge_out_projection = self.get_einsum_dense(
                     'ihj,jkh->ihk', (self.num_heads, self.units))
 
-        # Future implementation: Allow for non-shared weights?
-        # I.e., implement node_projection_dst and node_projection_src
-        self.node_projection = self.get_einsum_dense(
+        self.projection = self.get_einsum_dense(
             'ij,jkh->ihk', (self.num_heads, self.units))
 
+        self.node_projection = self.get_einsum_dense(
+            'ij,jkh->ihk', (self.num_heads, self.units))
+        
         if self.apply_self_projection:
             self.self_projection = self.get_einsum_dense(
                 'ij,jkh->ihk', (self.num_heads, self.units))
@@ -230,56 +202,52 @@ class GATv2Conv(BaseLayer):
         if self.merge_mode == 'concat':
             self.units *= self.num_heads
 
-    def subclass_call(self, tensor: GraphTensor) -> GraphTensor:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
 
+        if self.apply_self_projection:
+            node_feature_residual = self.self_projection(tensor.node_feature)
+        else:
+            node_feature_residual = None
+
+        # GAT v. GATv2: Concatenation before linear projection
+        if not self.use_edge_features:
+            attention_feature = tf.concat([
+                tf.gather(tensor.node_feature, tensor.edge_dst),
+                tf.gather(tensor.node_feature, tensor.edge_src)], axis=-1)
+        else:
+            attention_feature = tf.concat([
+                tf.gather(tensor.node_feature, tensor.edge_dst),
+                tf.gather(tensor.node_feature, tensor.edge_src),
+                tensor.edge_feature], axis=-1)
+        
         node_feature = self.node_projection(tensor.node_feature)
-        # Add dst and src node features, instead of concatenating them
-        attention_feature = (
-            tf.gather(node_feature, tensor.edge_dst) +
-            tf.gather(node_feature, tensor.edge_src)
-        )
-        if self.use_edge_features:
-            edge_feature = self.edge_projection(tensor.edge_feature)
-            # Similarily, add associated edge features, instead of concatenating them
-            attention_feature += edge_feature
 
+        attention_feature = self.projection(attention_feature)
+
+        if self.use_edge_features:
             if self.update_edge_features:
                 edge_feature = self.edge_out_projection(attention_feature)
-                edge_feature = reduce_features(
+                edge_feature = gnn_ops.reduce_features(
                     feature=edge_feature,
                     mode=self.merge_mode,
                     output_units=self.units)
-
                 tensor = tensor.update({'edge_feature': edge_feature})
 
-        # Apply activation before attention projection
+        # GAT v. GATv2: Activation before linear projection
         edge_weights = self.attention_activation(attention_feature)
         edge_weights = self.attention_projection(edge_weights)
 
-        edge_weights = softmax_edge_weights(
-            edge_weight=edge_weights,
-            edge_dst=tensor.edge_dst)
-
-        node_feature = propagate_node_features(
-            node_feature=node_feature,
-            edge_dst=tensor.edge_dst,
-            edge_src=tensor.edge_src,
-            edge_weight=edge_weights)
-
-        if self.apply_self_projection:
-            node_feature += self.self_projection(tensor.node_feature)
-
-        node_feature = reduce_features(
-            feature=node_feature,
-            mode=self.merge_mode,
-            output_units=self.units)
-
-        return tensor.update({'node_feature': node_feature})
+        tensor = tensor.update({
+            'node_feature': node_feature, 'edge_weight': edge_weights})
+        
+        return tensor.propagate(
+            normalize=True, 
+            reduction=self.merge_mode,
+            residual=node_feature_residual)
 
     def get_config(self):
         base_config = super().get_config()
         config = {
-            'use_edge_features': self.use_edge_features,
             'num_heads': self.num_heads,
             'merge_mode': self.merge_mode,
             'self_projection': self.apply_self_projection,
