@@ -1,77 +1,46 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
-from tensorflow.keras import layers
+
+from keras import initializers
+from keras import regularizers
+from keras import constraints
+from keras import layers
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import propagate_node_features
-from molgraph.layers.ops import reduce_features
+
+from molgraph.layers import gnn_layer
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class GMMConv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class GMMConv(gnn_layer.GNNLayer):
 
     '''Multi-head graph gaussian mixture layer (MoNet)
 
-    Implementation is based on Dwivedi et al. (2022) [#]_ and Monti et al. (2016) [#]_.
+    Implementation is based on Dwivedi et al. (2022) [#]_ and 
+    Monti et al. (2016) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GMMConv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GMMConv(16, activation='relu'),
-    ...     molgraph.layers.GMMConv(16, activation='relu')
+    ...     molgraph.layers.GMMConv(units=16),
+    ...     molgraph.layers.GMMConv(units=16),
+    ...     molgraph.layers.GMMConv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Build a model with GMMConv
-    >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GMMConv(16, activation='relu'),
-    ...     molgraph.layers.GMMConv(16, activation='relu')
-    ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
 
     Args:
         units (int, None):
@@ -86,8 +55,9 @@ class GMMConv(BaseLayer):
             Default to 2.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -113,6 +83,15 @@ class GMMConv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
+
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
 
     References:
 
@@ -128,17 +107,13 @@ class GMMConv(BaseLayer):
         merge_mode: Optional[str] = 'sum',
         pseudo_coord_dim=2,
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = 'relu',
         use_bias: bool = True,
-        kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.TruncatedNormal(stddev=0.005),
-        bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+        kernel_initializer: Union[str, initializers.Initializer, None] = None,
+        bias_initializer: Union[str, initializers.Initializer, None] = None,
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -148,7 +123,7 @@ class GMMConv(BaseLayer):
     ):
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -160,6 +135,7 @@ class GMMConv(BaseLayer):
             activity_regularizer=activity_regularizer,
             kernel_constraint=kernel_constraint,
             bias_constraint=bias_constraint,
+            use_edge_features=kwargs.pop('use_edge_features', False),
             **kwargs)
 
         self.num_kernels = num_kernels
@@ -167,11 +143,7 @@ class GMMConv(BaseLayer):
         self.pseudo_coord_dim = pseudo_coord_dim
         self.apply_self_projection = self_projection
 
-    def subclass_build(
-        self,
-        node_feature_shape: Optional[tf.TensorShape] = None,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
 
         if self.merge_mode == 'concat':
             if not self.units or (self.units % self.num_kernels != 0):
@@ -203,28 +175,24 @@ class GMMConv(BaseLayer):
         if self.merge_mode == 'concat':
             self.units *= self.num_kernels
 
-    def subclass_call(self, tensor: GraphTensor) -> GraphTensor:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
+
+        if self.apply_self_projection:
+            node_feature_residual = self.self_projection(tensor.node_feature)
+        else:
+            node_feature_residual = None
 
         node_feature = self.node_projection(tensor.node_feature)
 
         edge_weights = self.compute_edge_weights(
             tensor.edge_dst, tensor.edge_src)
 
-        node_feature = propagate_node_features(
-            node_feature=node_feature,
-            edge_dst=tensor.edge_dst,
-            edge_src=tensor.edge_src,
-            edge_weight=edge_weights)
+        tensor = tensor.update({
+            'node_feature': node_feature, 'edge_weight': edge_weights})
 
-        if self.apply_self_projection:
-            node_feature += self.self_projection(tensor.node_feature)
-
-        node_feature = reduce_features(
-            feature=node_feature,
-            mode=self.merge_mode,
-            output_units=self.units)
-
-        return tensor.update({'node_feature': node_feature})
+        return tensor.propagate(
+            reduction=self.merge_mode,
+            residual=node_feature_residual)
 
     def compute_edge_weights(self, edge_dst, edge_src, clip_values=(-5, 5)):
         """Computes edge weights via Gaussian kernels"""
@@ -236,8 +204,9 @@ class GMMConv(BaseLayer):
             degree = tf.cast(degree, tf.float32)
             degree = tf.where(degree == 0.0, 1.0, degree) ** -(1/2)
             degree = tf.stack([
+                tf.gather(degree, edge_src),
                 tf.gather(degree, edge_dst),
-                tf.gather(degree, edge_src)], axis=1)
+            ], axis=1)
 
             # (n_edges, 2) @ (2, 2) -> (n_edges, 2)
             pseudo_coord = self.pseudo_coord_projection(degree)
