@@ -1,77 +1,46 @@
 import tensorflow as tf
-from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
+from tensorflow import keras 
+
+from keras import initializers
+from keras import regularizers
+from keras import constraints
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import compute_edge_weights_from_degrees
-from molgraph.layers.ops import propagate_node_features
+
+from molgraph.layers import gnn_layer
+from molgraph.layers import gnn_ops
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class GCNConv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class GCNConv(gnn_layer.GNNLayer):
 
     """Graph convolutional layer (GCN).
 
     Implementation is based on Kipf et al. (2017) [#]_, Dwivedi et al. (2022) [#]_,
     and, for RGCN, Schlichtkrull et al. (2017) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GCNConv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GCNConv(16, activation='relu'),
-    ...     molgraph.layers.GCNConv(16, activation='relu')
+    ...     molgraph.layers.GCNConv(units=16),
+    ...     molgraph.layers.GCNConv(units=16),
+    ...     molgraph.layers.GCNConv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Build a model with GCNConv
-    >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GCNConv(16, activation='relu'),
-    ...     molgraph.layers.GCNConv(16, activation='relu')
-    ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
 
     Args:
         units (int, None):
@@ -86,8 +55,9 @@ class GCNConv(BaseLayer):
             use_edge_features is True. Default to None.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -113,6 +83,18 @@ class GCNConv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
+
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
+            *   `use_edge_features`: Whether or not to use edge features. 
+                Only relevant if edge features exist. Default to False.
+
 
     References:
         .. [#] https://arxiv.org/pdf/1609.02907.pdf
@@ -123,21 +105,16 @@ class GCNConv(BaseLayer):
     def __init__(
         self,
         units: Optional[int] = None,
-        use_edge_features: bool = False,
         degree_normalization: Optional[str] = 'symmetric',
         num_bases: Optional[int] = None,
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = 'relu',
         use_bias: bool = True,
-        kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.TruncatedNormal(stddev=0.005),
-        bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+        kernel_initializer: Union[str, initializers.Initializer, None] = None,
+        bias_initializer: Union[str, initializers.Initializer, None] = None,
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -145,10 +122,9 @@ class GCNConv(BaseLayer):
         bias_constraint: Optional[constraints.Constraint] = None,
         **kwargs
     ):
-        kwargs.pop("use_bias", None)
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -160,30 +136,24 @@ class GCNConv(BaseLayer):
             activity_regularizer=activity_regularizer,
             kernel_constraint=kernel_constraint,
             bias_constraint=bias_constraint,
+            use_edge_features=kwargs.pop('use_edge_features', False),
             **kwargs)
 
-        self.use_edge_features = use_edge_features
         self.degree_normalization = degree_normalization
         self.num_bases = (
             int(num_bases) if isinstance(num_bases, (int, float)) else 0)
         self.apply_self_projection = self_projection
         self.use_bias = use_bias
 
-    def subclass_build(
-        self,
-        node_feature_shape: Optional[tf.TensorShape] = None,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
 
         kernel_shape = [self.node_dim, self.units]
         bias_shape = [self.units]
 
-        if edge_feature_shape is None:
-            self.use_edge_features = False
 
         if self.use_edge_features:
 
-            edge_dim = edge_feature_shape[-1]
+            edge_dim = graph_tensor_spec.edge_feature.shape[-1]
 
             if self.num_bases > 0 and self.num_bases < edge_dim:
                 num_bases = self.num_bases
@@ -203,25 +173,23 @@ class GCNConv(BaseLayer):
         if self.apply_self_projection:
             self.self_projection = self.get_dense(self.units)
 
-    def subclass_call(self, tensor: GraphTensor) -> GraphTensor:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
 
         if self.use_edge_features:
             node_feature = tf.expand_dims(tensor.node_feature, axis=-1)
         else:
             node_feature = tensor.node_feature
 
-        edge_weight = compute_edge_weights_from_degrees(
-            edge_dst=tensor.edge_dst,
+        edge_weight = gnn_ops.compute_edge_weights_from_degrees(
             edge_src=tensor.edge_src,
+            edge_dst=tensor.edge_dst,
             edge_feature=tensor.edge_feature if self.use_edge_features else None,
             mode=self.degree_normalization)
 
         # (n_edges, ndim, 1) x (n_edges, 1, edim) -> (n_edges, ndim, edim)
-        node_feature = propagate_node_features(
-            node_feature=node_feature,
-            edge_dst=tensor.edge_dst,
-            edge_src=tensor.edge_src,
-            edge_weight=edge_weight)
+        tensor_update = tensor.update({
+            'node_feature': node_feature, 'edge_weight': edge_weight})
+        node_feature = tensor_update.propagate().node_feature
 
         if hasattr(self, 'kernel_decomp'):
             # (n_dim, unit, n_bases) x (e_dim, n_bases) -> (n_dim, unit, e_dim)
@@ -247,7 +215,6 @@ class GCNConv(BaseLayer):
     def get_config(self):
         base_config = super().get_config()
         config = {
-            'use_edge_features': self.use_edge_features,
             'self_projection': self.apply_self_projection,
             'degree_normalization': self.degree_normalization,
             'num_bases': self.num_bases,
