@@ -1,76 +1,45 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
+
+from keras import initializers
+from keras import regularizers
+from keras import constraints
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import compute_edge_weights_from_degrees
-from molgraph.layers.ops import propagate_node_features
+
+from molgraph.layers import gnn_layer
+from molgraph.layers import gnn_ops
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class GCNIIConv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class GCNIIConv(gnn_layer.GNNLayer):
 
     '''Graph convolutional 'via Initial residual and Identity mapping' layer (GCNII).
 
     Implementation is based on Chen et al. (2020) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GCNIIConv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GCNIIConv(16, activation='relu'),
-    ...     molgraph.layers.GCNIIConv(16, activation='relu')
+    ...     molgraph.layers.GCNIIConv(units=16),
+    ...     molgraph.layers.GCNIIConv(units=16),
+    ...     molgraph.layers.GCNIIConv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Build a model with GCNIIConv
-    >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GCNIIConv(16, activation='relu'),
-    ...     molgraph.layers.GCNIIConv(16, activation='relu')
-    ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
 
     Args:
         units (int, None):
@@ -95,8 +64,9 @@ class GCNIIConv(BaseLayer):
             'symmetric', 'row' or None. If None, 'row' is used. Default to 'symmetric'.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -122,6 +92,16 @@ class GCNIIConv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
+
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
+
 
     References:
         .. [#] https://arxiv.org/pdf/2007.02133v1.pdf
@@ -136,17 +116,13 @@ class GCNIIConv(BaseLayer):
         variant: bool = False,
         degree_normalization: str = 'symmetric',
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = 'relu',
         use_bias: bool = True,
-        kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.TruncatedNormal(stddev=0.005),
-        bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+        kernel_initializer: Union[str, initializers.Initializer, None] = None,
+        bias_initializer: Union[str, initializers.Initializer, None] = None,
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -156,7 +132,7 @@ class GCNIIConv(BaseLayer):
     ):
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -168,6 +144,7 @@ class GCNIIConv(BaseLayer):
             activity_regularizer=activity_regularizer,
             kernel_constraint=kernel_constraint,
             bias_constraint=bias_constraint,
+            use_edge_features=kwargs.pop('use_edge_features', False),
             **kwargs)
 
         self.alpha = alpha
@@ -177,43 +154,37 @@ class GCNIIConv(BaseLayer):
         self.apply_self_projection = self_projection
         self.residual = residual
 
-    def subclass_build(
-        self,
-        node_feature_shape: tf.TensorShape,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
 
         self.projection = self.get_dense(self.units)
 
-        if (
-            self.units != node_feature_shape[-1] and
-            not hasattr(self, 'node_resample')
-        ):
+        node_dim = graph_tensor_spec.node_feature.shape[-1]
+        
+        if self.units != node_dim and not hasattr(self, 'node_resample'):
             self.node_resample = self.get_dense(self.units)
 
         if self.apply_self_projection:
             self.self_projection = self.get_dense(self.units)
 
-    def subclass_call(self, tensor: GraphTensor) -> None:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
 
         if hasattr(self, 'node_resample'):
             tensor = tensor.update({
                 'node_feature': self.node_resample(tensor.node_feature)})
 
         if not hasattr(tensor, 'node_feature_initial'):
-            tensor = tensor.update({'node_feature_initial': tensor.node_feature})
+            tensor = tensor.update({
+                'node_feature_initial': tensor.node_feature})
 
-        edge_weight = compute_edge_weights_from_degrees(
-            edge_dst=tensor.edge_dst,
+        edge_weight = gnn_ops.compute_edge_weights_from_degrees(
             edge_src=tensor.edge_src,
+            edge_dst=tensor.edge_dst,
             edge_feature=None,
             mode=self.degree_normalization)
 
-        node_feature = propagate_node_features(
-            node_feature=tensor.node_feature,
-            edge_dst=tensor.edge_dst,
-            edge_src=tensor.edge_src,
-            edge_weight=edge_weight)
+        tensor = tensor.update({'edge_weight': edge_weight})
+        
+        node_feature = tensor.propagate().node_feature
 
         identity = (
             (1 - self.alpha) * node_feature +
@@ -223,7 +194,7 @@ class GCNIIConv(BaseLayer):
         if self.variant:
             node_feature = tf.concat([
                 node_feature * (1 - self.alpha),
-                tensor.node_features_initial * self.alpha
+                tensor.node_feature_initial * self.alpha
             ], axis=1)
         else:
             node_feature = identity
