@@ -1,78 +1,46 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
-from tensorflow.keras import layers
+from keras import initializers
+from keras import regularizers
+from keras import constraints
+from keras import activations
+from keras import layers
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import compute_edge_weights_from_degrees
-from molgraph.layers.ops import propagate_node_features
+
+from molgraph.layers import gnn_layer
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class GraphSageConv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class GraphSageConv(gnn_layer.GNNLayer):
 
     '''Graph sage convolution layer (GraphSage)
 
     Implementation is based on Hamilton et al. (2018) [#]_ and
     Dwivedi et al. (2022) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Build a model with GraphSageConv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GraphSageConv(16, activation='relu'),
-    ...     molgraph.layers.GraphSageConv(16, activation='relu')
+    ...     molgraph.layers.GraphSageConv(units=16),
+    ...     molgraph.layers.GraphSageConv(units=16),
+    ...     molgraph.layers.GraphSageConv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Build a model with GraphSageConv
-    >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.GraphSageConv(16, activation='relu'),
-    ...     molgraph.layers.GraphSageConv(16, activation='relu')
-    ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
 
     Args:
         units (int, None):
@@ -85,8 +53,9 @@ class GraphSageConv(BaseLayer):
             features. Default to True.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -112,6 +81,15 @@ class GraphSageConv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
+
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
 
     References:
         .. [#] https://arxiv.org/pdf/1706.02216.pdf
@@ -125,17 +103,13 @@ class GraphSageConv(BaseLayer):
         aggregation_mode: str = 'mean',
         normalize: bool = True,
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = 'relu',
         use_bias: bool = True,
-        kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.TruncatedNormal(stddev=0.005),
-        bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+        kernel_initializer: Union[str, initializers.Initializer, None] = None,
+        bias_initializer: Union[str, initializers.Initializer, None] = None,
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -145,7 +119,7 @@ class GraphSageConv(BaseLayer):
     ):
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -157,18 +131,15 @@ class GraphSageConv(BaseLayer):
             activity_regularizer=activity_regularizer,
             kernel_constraint=kernel_constraint,
             bias_constraint=bias_constraint,
+            use_edge_features=kwargs.pop('use_edge_features', False),
             **kwargs)
 
         self.aggregation_mode = aggregation_mode
-        self.normalize = normalize if not batch_norm else False
+        self.normalize = normalize if not normalization else False
         self.apply_self_projection = self_projection
         self.activation = activations.get('relu')
 
-    def subclass_build(
-        self,
-        node_feature_shape: Optional[tf.TensorShape] = None,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
 
         if self.aggregation_mode == 'max':
             self.node_src_projection = self.get_dense(self.units)
@@ -180,32 +151,30 @@ class GraphSageConv(BaseLayer):
 
         self.node_projection = self.get_dense(self.units)
 
-    def subclass_call(self, tensor: GraphTensor) -> GraphTensor:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
+        
+        node_feature_residual = tensor.node_feature
 
-        if self.aggregation_mode == 'max':
-            node_feature = self.node_src_projection(tensor.node_feature)
-            node_feature = self.activation(node_feature)
-            node_feature = propagate_node_features(
-                node_feature=node_feature,
-                edge_dst=tensor.edge_dst,
-                edge_src=tensor.edge_src,
-                mode=self.aggregation_mode)
-        elif self.aggregation_mode == 'lstm':
+        if self.aggregation_mode == 'lstm':
             node_feature = self.lstm_aggregate(
-                tensor.node_feature, tensor.edge_dst, tensor.edge_src)
+                tensor.node_feature, tensor.edge_src, tensor.edge_dst)
         else:
-            node_feature = propagate_node_features(
-                node_feature=tensor.node_feature,
-                edge_dst=tensor.edge_dst,
-                edge_src=tensor.edge_src,
-                mode=self.aggregation_mode)
+            if self.aggregation_mode == 'max':
+                node_feature = self.node_src_projection(tensor.node_feature)
+                node_feature = self.activation(node_feature)
+            else:
+                node_feature = tensor.node_feature
 
-        node_feature = tf.concat([node_feature, tensor.node_feature], axis=-1)
+            tensor = tensor.update({'node_feature': node_feature})
+            tensor = tensor.propagate(mode=self.aggregation_mode)
+            
+        node_feature = tf.concat([
+            tensor.node_feature, node_feature_residual], axis=-1)
 
         node_feature = self.node_projection(node_feature)
 
         if self.apply_self_projection:
-            node_feature += self.self_projection(tensor.node_feature)
+            node_feature += self.self_projection(node_feature_residual)
 
         if self.normalize:
             node_feature = self.activation(node_feature)
@@ -213,38 +182,30 @@ class GraphSageConv(BaseLayer):
 
         return tensor.update({'node_feature': node_feature})
 
-    def lstm_aggregate(self, node_feature, edge_dst, edge_src):
+    def lstm_aggregate(self, node_feature, edge_src, edge_dst):
 
-        def true_fn(node_feature, edge_dst, edge_src):
-            """If edges exist, call this function"""
+        def true_fn(node_feature, edge_src, edge_dst):
+            'If edges exist, call this function'
 
-            node_indices = tf.unique(edge_dst)[0]
+            # Get number of nodes
             num_nodes = tf.shape(node_feature)[0]
-            #print(node_indices)
 
-            # A bit of a hack to shuffle neighbor (source) nodes of the
-            # destination nodes.
-            random_indices = tf.random.shuffle(tf.range(tf.shape(edge_dst)[0]))
+            # Shuffle neighbor (source) nodes of the destination nodes.
+            random_indices = tf.random.shuffle(tf.range(tf.shape(edge_src)[0]))
             edge_dst = tf.gather(edge_dst, random_indices)
             edge_src = tf.gather(edge_src, random_indices)
             sorted_indices = tf.argsort(edge_dst)
             edge_dst = tf.gather(edge_dst, sorted_indices)
-            edge_dst -= tf.reduce_min(edge_dst)
             edge_src = tf.gather(edge_src, sorted_indices)
 
             # Gather source nodes followed by a partitioning of destination nodes
             node_feature = tf.RaggedTensor.from_value_rowids(
-                tf.gather(node_feature, edge_src), edge_dst)
+                tf.gather(node_feature, edge_src), edge_dst, num_nodes)
             node_feature = node_feature.to_tensor()
             # Pass to lstm for update
             node_feature = self.lstm(node_feature)
 
-            node_feature_dim = tf.shape(node_feature)[-1]
-
-            return tf.scatter_nd(
-                indices=tf.expand_dims(node_indices, axis=-1),
-                updates=node_feature,
-                shape=(num_nodes, node_feature_dim))
+            return node_feature
 
         def false_fn(node_feature):
             """If no edges exist, call this function"""
@@ -253,8 +214,8 @@ class GraphSageConv(BaseLayer):
                 dtype=node_feature.dtype)
 
         return tf.cond(
-            tf.greater(tf.shape(edge_dst)[0], 0),
-            lambda: true_fn(node_feature, edge_dst, edge_src),
+            tf.greater(tf.shape(edge_src)[0], 0),
+            lambda: true_fn(node_feature, edge_src, edge_dst),
             lambda: false_fn(node_feature)
         )
 
