@@ -1,82 +1,48 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
 
-import numpy as np
+from keras import initializers
+from keras import regularizers
+from keras import constraints
 
 from typing import Optional
 from typing import Callable
 from typing import Union
-from typing import Tuple
+
+from molgraph.internal import register_keras_serializable 
 
 from molgraph.tensors.graph_tensor import GraphTensor
-from molgraph.layers.base import BaseLayer
-from molgraph.layers.ops import propagate_node_features
-from molgraph.layers.geometric import _radial_basis
+
+from molgraph.layers import gnn_layer
+
+from molgraph.layers.geometric import radial_basis
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
-class DTNNConv(BaseLayer):
+@register_keras_serializable(package='molgraph')
+class DTNNConv(gnn_layer.GNNLayer):
 
     """Deep Tensor Neural Network (DTNN).
 
     Implementation is based on Schütt et al. (2017a) [#]_.
 
-    **Examples:**
-
-    Inputs a ``GraphTensor`` encoding (two) subgraphs:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...         # edge_feature encodes distances between edge_dst and edge_src
-    ...         'edge_feature': [[0.3, 0.3], [0.1, 0.2, 0.1, 0.4, 0.4, 0.2]],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
+    ...     # edge_feature encodes e.g. distances between edge_src and edge_dst
+    ...     edge_feature=[0.3, 0.3, 0.1, 0.2, 0.1, 0.4, 0.4, 0.2],
     ... )
-    >>> # Build a model with DTNNConv
     >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.DTNNConv(16, activation='relu'),
-    ...     molgraph.layers.DTNNConv(16, activation='relu')
+    ...     molgraph.layers.DTNNConv(units=16),
+    ...     molgraph.layers.DTNNConv(units=16),
+    ...     molgraph.layers.DTNNConv(units=16),
+    ...     molgraph.layers.Readout(),
     ... ])
-    >>> gnn_model.output_shape
-    (None, None, 16)
-
-    Inputs a ``GraphTensor`` encoding a single disjoint graph:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [1.0, 0.0],
-    ...             [0.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...         # edge_feature encodes distances between edge_dst and edge_src
-    ...         'edge_feature': [0.3, 0.3, 0.1, 0.2, 0.1, 0.4, 0.4, 0.2],
-    ...     }
-    ... )
-    >>> # Build a model with DTNNConv
-    >>> gnn_model = tf.keras.Sequential([
-    ...     tf.keras.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.DTNNConv(16, activation='relu'),
-    ...     molgraph.layers.DTNNConv(16, activation='relu')
-    ... ])
-    >>> gnn_model.output_shape
-    (None, 16)
+    >>> gnn_model(graph_tensor).shape
+    TensorShape([2, 16])
 
     Args:
         units (int, None):
@@ -101,8 +67,9 @@ class DTNNConv(BaseLayer):
             Default to 'auto'.
         self_projection (bool):
             Whether to apply self projection. Default to True.
-        batch_norm: (bool):
-            Whether to apply batch normalization to the output. Default to True.
+        normalization: (None, str, bool):
+            Whether to apply layer normalization to the output. If batch 
+            normalization is desired, pass 'batch_norm'. Default to None.
         residual: (bool)
             Whether to add skip connection to the output. Default to True.
         dropout: (float, None):
@@ -128,7 +95,16 @@ class DTNNConv(BaseLayer):
             Constraint function applied to the kernels. Default to None.
         bias_constraint (tf.keras.constraints.Constraint, None):
             Constraint function applied to the biases. Default to None.
+        **kwargs: Valid (optional) keyword arguments are:
 
+            *   `name` (str): Name of the layer instance.
+            *   `update_step` (tf.keras.layers.Layer): Applies post-processing 
+                step on the output (produced by `_call`). If passed, 
+                `normalization`, `residual`, `activation` and `dropout` 
+                parameters will be ignored. If None, a default post-processing 
+                step will be used (taking into consideration the aforementioned 
+                parameters). Default to None.
+                
     References:
         .. [#] https://www.nature.com/articles/ncomms13890
     """
@@ -142,17 +118,15 @@ class DTNNConv(BaseLayer):
         distance_granularity: float = 0.1,
         rbf_stddev: Optional[Union[float, str]] = 'auto',
         self_projection: bool = True,
-        batch_norm: bool = True,
+        normalization: Union[None, str, bool] = None,
         residual: bool = True,
         dropout: Optional[float] = None,
         activation: Union[None, str, Callable[[tf.Tensor], tf.Tensor]] = None,
         use_bias: bool = True,
         kernel_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.GlorotUniform(),
+            str, initializers.Initializer, None] = 'glorot_uniform',
         bias_initializer: Union[
-            str, initializers.Initializer
-        ] = initializers.Constant(0.),
+            str, initializers.Initializer, None] = 'zeros',
         kernel_regularizer: Optional[regularizers.Regularizer] = None,
         bias_regularizer: Optional[regularizers.Regularizer] = None,
         activity_regularizer: Optional[regularizers.Regularizer] = None,
@@ -162,7 +136,7 @@ class DTNNConv(BaseLayer):
     ):
         super().__init__(
             units=units,
-            batch_norm=batch_norm,
+            normalization=normalization,
             residual=residual,
             dropout=dropout,
             activation=activation,
@@ -174,6 +148,7 @@ class DTNNConv(BaseLayer):
             activity_regularizer=activity_regularizer,
             kernel_constraint=kernel_constraint,
             bias_constraint=bias_constraint,
+            use_edge_features=kwargs.pop('use_edge_features', True),
             **kwargs)
 
         self.apply_self_projection = self_projection
@@ -182,18 +157,15 @@ class DTNNConv(BaseLayer):
         self.distance_granularity = distance_granularity
         self.rbf_stddev = rbf_stddev
 
-        self.rbf = _radial_basis.RadialBasis(
+        self.rbf = radial_basis.RadialBasis(
             self.distance_min,
             self.distance_max,
             self.distance_granularity,
             self.rbf_stddev
         )
 
-    def subclass_build(
-        self,
-        node_feature_shape: Optional[tf.TensorShape] = None,
-        edge_feature_shape: Optional[tf.TensorShape] = None
-    ) -> None:
+    def _build(self, graph_tensor_spec: GraphTensor.Spec) -> None:
+        
         self.edge_projection = self.get_dense(self.units)
         self.node_projection_1 = self.get_dense(self.units)
         self.node_projection_2 = self.get_dense(self.units)
@@ -201,7 +173,7 @@ class DTNNConv(BaseLayer):
         if self.apply_self_projection:
             self.self_projection = self.get_dense(self.units)
 
-    def subclass_call(self, tensor: GraphTensor) -> GraphTensor:
+    def _call(self, tensor: GraphTensor) -> GraphTensor:
 
         num_segments = tf.shape(tensor.node_feature)[0]
 
