@@ -1,44 +1,37 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import layers
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
+from keras import layers
+from keras import initializers
+from keras import regularizers
+from keras import constraints
+from keras import activations
 
 from typing import Union
 from typing import Callable
 from typing import Optional
 
+from molgraph.internal import register_keras_serializable 
+
 from molgraph.tensors.graph_tensor import GraphTensor
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class LaplacianPositionalEncoding(layers.Layer):
 
     '''Laplacian positional encoding.
 
     Implementation based on Dwivedi et al. (2021) [#]_ and Belkin et al. (2003) [#]_.
 
-    **Example:**
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [1.0, 1.0],
-    ...             [1.0, 1.0],
-    ...             [1.0, 1.0],
-    ...             [1.0, 1.0],
-    ...             [1.0, 1.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
     >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     molgraph.layers.LaplacianPositionalEncoding(16)
+    ...     molgraph.layers.LaplacianPositionalEncoding(16),
     ... ])
     >>> graph_tensor = model(graph_tensor)
     >>> graph_tensor.node_feature != 1.0
@@ -148,7 +141,7 @@ class LaplacianPositionalEncoding(layers.Layer):
             kernel_constraint=self.kernel_constraint,
             bias_constraint=self.bias_constraint)
 
-    def call(self, tensor: GraphTensor) -> GraphTensor:
+    def call(self, tensor: GraphTensor, training: Optional[bool] = None) -> GraphTensor:
         '''Defines the computation from inputs to outputs.
 
         This method should not be called directly, but indirectly
@@ -170,7 +163,7 @@ class LaplacianPositionalEncoding(layers.Layer):
         if not self._built:
             self._build(
                 getattr(tensor, 'node_feature', None),
-                getattr(tensor, 'positional_encoding', None)
+                getattr(tensor, 'node_position', None)
             )
 
         def random_sign_flip(positional_encoding):
@@ -182,9 +175,10 @@ class LaplacianPositionalEncoding(layers.Layer):
             positional_encoding = compute_positional_encoding(
                 tensor, self._target_dim)
         else:
-            positional_encoding = tensor.positional_encoding
+            positional_encoding = tensor.node_position
 
-        positional_encoding = random_sign_flip(positional_encoding)
+        if training:
+            positional_encoding = random_sign_flip(positional_encoding)
         node_feature = tensor.node_feature + self.projection(positional_encoding)
         return tensor_orig.update({'node_feature': node_feature})
 
@@ -204,12 +198,12 @@ class LaplacianPositionalEncoding(layers.Layer):
     def get_config(self):
         base_config = super().get_config()
         config = {
-            'dim': self.dim,
+            'dim': self._target_dim,
             'activation': activations.serialize(self.activation),
             'use_bias': self.use_bias,
             'kernel_initializer': initializers.serialize(self.kernel_initializer),
             'bias_initializer': initializers.serialize(self.bias_initializer),
-            'kernel_regularizer': regularizers.serialize(sself.kernel_regularizer),
+            'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
             'bias_regularizer': regularizers.serialize(self.bias_regularizer),
             'activity_regularizer': regularizers.serialize(self.activity_regularizer),
             'kernel_constraint': constraints.serialize(self.kernel_constraint),
@@ -222,7 +216,8 @@ class LaplacianPositionalEncoding(layers.Layer):
 
 
 def compute_normalized_laplacian(adjacency, num_nodes):
-    degree = tf.math.bincount(adjacency[:, 0])
+    degree = tf.math.bincount(
+        tf.cast(adjacency[:, 1], tf.int32), dtype=adjacency.dtype)
     degree = tf.gather(degree, adjacency)
     adjacency_norm = tf.reduce_prod(tf.cast(degree, tf.float32), 1) ** -0.5
     laplacian_norm = tf.scatter_nd(
@@ -242,7 +237,7 @@ def compute_eigen_vectors(laplacian, target_dim):
 def compute_positional_encoding(tensor, target_dim):
 
     num_nodes = tf.shape(tensor.node_feature)[0]
-    adjacency = tf.stack([tensor.edge_dst, tensor.edge_src], axis=1)
+    adjacency = tf.stack([tensor.edge_src, tensor.edge_dst], axis=1)
 
     laplacian_norm = compute_normalized_laplacian(adjacency, num_nodes)
 
