@@ -1,42 +1,38 @@
 import tensorflow as tf
 from tensorflow import keras
 
+from molgraph.internal import register_keras_serializable 
+
 from molgraph.tensors.graph_tensor import GraphTensor
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class GatherIncident(keras.layers.Layer):
     '''Gathers incident node features.
 
     Useful for e.g., downstream edge and link classification.
 
-    **Example:**
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [2.0, 0.0]],
-    ...             [[3.0, 0.0], [4.0, 0.0], [0.0, 5.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> graph_tensor = graph_tensor.merge()
     >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
     ...     molgraph.layers.GatherIncident(concat=True)
     ... ])
     >>> model(graph_tensor)
     <tf.Tensor: shape=(8, 4), dtype=float32, numpy=
-    array([[1., 0., 2., 0.],
-           [2., 0., 1., 0.],
-           [3., 0., 4., 0.],
-           [3., 0., 0., 5.],
-           [4., 0., 3., 0.],
-           [4., 0., 0., 5.],
-           [0., 5., 4., 0.],
-           [0., 5., 3., 0.]], dtype=float32)>
+    array([[1., 0., 1., 0.],
+           [1., 0., 1., 0.],
+           [1., 0., 1., 0.],
+           [0., 1., 1., 0.],
+           [1., 0., 1., 0.],
+           [0., 1., 1., 0.],
+           [1., 0., 0., 1.],
+           [1., 0., 0., 1.]], dtype=float32)>
 
     Args:
         concat (bool):
@@ -71,13 +67,13 @@ class GatherIncident(keras.layers.Layer):
         node_feature_dst = tf.gather(tensor.node_feature, tensor.edge_dst)
         if self.concat:
             node_feature_incident = tf.concat([
-                node_feature_dst, node_feature_src], axis=1)
+                node_feature_src, node_feature_dst], axis=1)
         else:
             node_feature_incident = tf.stack([
-                node_feature_dst, node_feature_src], axis=1)
+                node_feature_src, node_feature_dst], axis=1)
         tensor_orig = tensor_orig.update({
-            'node_feature_incident': node_feature_incident})
-        return tensor_orig.node_feature_incident
+            'edge_feature_incident': node_feature_incident})
+        return tensor_orig.edge_feature_incident
 
     def get_config(self):
         base_config = super().get_config()
