@@ -1,84 +1,67 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import layers
-from tensorflow.keras import initializers
-from tensorflow.keras import regularizers
-from tensorflow.keras import constraints
-from tensorflow.keras import activations
+from keras import layers
+from keras import initializers
+from keras import regularizers
+from keras import constraints
 
 from typing import Optional
 from typing import Union
 from typing import List
 
+from molgraph.internal import register_keras_serializable 
+
 from molgraph.tensors.graph_tensor import GraphTensor
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class EmbeddingLookup(layers.StringLookup):
 
     '''A loookup layer and embedding layer in combination.
 
     Specify, as keyword argument only,
-    ``EmbeddingLookup(feature='node_feature')`` to perform standard scaling
-    on the ``node_feature`` field of the ``GraphTensor``, or,
-    ``EmbeddingLookup(feature='edge_feature')`` to perform standard scaling
-    on the ``edge_feature`` field of the ``GraphTensor``. If not specified,
-    the ``node_feature`` field will be considered.
+    ``EmbeddingLookup(feature='node_feature', ...)`` to perform embedding lookup
+    on the ``node_feature`` component of the ``GraphTensor``, or,
+    ``EmbeddingLookup(feature='edge_feature', ...)`` to perform embedding lookup
+    on the ``edge_feature`` component of the ``GraphTensor``. If not specified,
+    the ``node_feature`` component will be considered.
 
-    **Examples:**
+    Instead of specifying `feature`, ``NodeEmbeddingLookup(...)`` or 
+    ``EdgeEmbeddingLookup(...)`` can be used instead.
 
-    Adapt layer on ``GraphTensor`` directly:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             'Sym:C', 'Sym:C', 'Sym:C', 'Sym:O', 'Sym:N',
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=['Sym:C', 'Sym:C', 'Sym:C', 'Sym:O', 'Sym:N'],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Initialize layer
     >>> embedding = molgraph.layers.EmbeddingLookup(
     ...    feature='node_feature', output_dim=4)
-    >>> # Adapt layer to graph_tensor
     >>> embedding.adapt(graph_tensor)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     embedding,
-    ... ])
+    >>> model = tf.keras.Sequential([embedding])
     >>> graph_tensor = model(graph_tensor)
     >>> graph_tensor.node_feature.shape
     TensorShape([5, 4])
 
-    Adapt layer on dataset, constructed from ``GraphTensor``:
+    Adapt layer on a tf.data.Dataset:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [2, 2, 3, 3, 4, 4]],
-    ...         'edge_src': [[1, 0], [3, 4, 2, 4, 3, 2]],
-    ...         'node_feature': [
-    ...             ['Sym:C', 'Sym:C'], ['Sym:C', 'Sym:O', 'Sym:N'],
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=['Sym:C', 'Sym:C', 'Sym:C', 'Sym:O', 'Sym:N'],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Obtain dataset
     >>> ds = tf.data.Dataset.from_tensor_slices(graph_tensor).batch(2)
-    >>> # Initialize layer
     >>> embedding = molgraph.layers.EmbeddingLookup(
     ...    feature='node_feature', output_dim=4)
-    >>> # Adapt layer to graph_tensor
     >>> embedding.adapt(ds)
-    >>> # Build model
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     embedding,
-    ... ])
-    >>> # Predict (obtain new GraphTensor)
-    >>> output = model.predict(ds)
+    >>> model = tf.keras.Sequential([embedding])
+    >>> output = model.predict(ds, verbose=0)
     >>> output.node_feature.shape
-    TensorShape([2, None, 4])
+    TensorShape([5, 4])
+
 
     Args:
         output_dim (int):
@@ -142,8 +125,6 @@ class EmbeddingLookup(layers.StringLookup):
         self.embeddings_initializer = initializers.get(embeddings_initializer)
         self.embeddings_regularizer = regularizers.get(embeddings_regularizer)
         self.embeddings_constraint = constraints.get(embeddings_constraint)
-        self._vocabulary_size = None
-        self._built_from_vocabulary_size = False
 
     def adapt(self, data, batch_size=None, steps=None):
         '''Adapts the layer to data.
@@ -163,13 +144,19 @@ class EmbeddingLookup(layers.StringLookup):
                 The number of steps of adaption. If None, the number of
                 samples divided by the batch_size is used. Default to None.
         '''
-        if not isinstance(data,  GraphTensor):
-            data = data = data.map(
+
+        if not isinstance(data, GraphTensor):
+            data = data.map(
                 lambda x: getattr(x, self.feature))
+            for x in data.take(1):
+                super().build(x.shape)
         else:
             data = getattr(data, self.feature)
+            super().build(data.shape)
+
         super().adapt(data, batch_size=batch_size, steps=steps)
-        self._vocabulary_size = self.vocabulary_size()
+
+        self._add_embedding()
 
     def call(self, tensor: GraphTensor) -> GraphTensor:
         '''Defines the computation from inputs to outputs.
@@ -185,11 +172,12 @@ class EmbeddingLookup(layers.StringLookup):
         Returns:
             GraphTensor:
                 A ``GraphTensor`` with updated features. Either the
-                ``node_features`` field or the ``edge_features``
-                field (of the ``GraphTensor``) are updated.
+                ``node_feature`` component or the ``edge_feature``
+                component (of the ``GraphTensor``) are updated.
         '''
-        if not self._built_from_vocabulary_size:
-            self._build_from_vocabulary_size(self._vocabulary_size)
+        if not hasattr(self, 'embeddings'):
+            with tf.init_scope():
+                self._add_embedding()
 
         tensor = tensor.update({
             self.feature: super().call(getattr(tensor, self.feature))
@@ -199,33 +187,16 @@ class EmbeddingLookup(layers.StringLookup):
                 self.embeddings, getattr(tensor, self.feature))
         })
 
-    def _build_from_vocabulary_size(self, vocabulary_size):
-        self._built_from_vocabulary_size = True
-
-        self.embeddings = self.add_weight(
-            shape=(vocabulary_size, self.output_dim),
-            dtype=tf.float32,
-            initializer=self.embeddings_initializer,
-            name='embeddings',
-            regularizer=self.embeddings_regularizer,
-            constraint=self.embeddings_constraint,
-            experimental_autocast=False
-        )
-
-    @classmethod
-    def from_config(cls, config):
-        vocabulary_size = config.pop('vocabulary_size')
-        layer = cls(**config)
-        if vocabulary_size is None:
-            pass
-        else:
-            layer._build_from_vocabulary_size(vocabulary_size)
-        return layer
+    def compute_output_shape(
+        self, 
+        input_shape: tf.TensorShape
+    ) -> tf.TensorShape:
+        return tf.TensorShape(
+            input_shape[:-1]).concatenate([self.output_dim])
 
     def get_config(self):
         base_config = super().get_config()
         base_config.update({
-            'vocabulary_size': self._vocabulary_size,
             'output_dim': self.output_dim,
             'embeddings_initializer': initializers.serialize(
                 self.embeddings_initializer),
@@ -236,12 +207,22 @@ class EmbeddingLookup(layers.StringLookup):
         })
         return base_config
 
+    def _add_embedding(self) -> None:
+        self.embeddings = self.add_weight(
+            shape=(self.vocabulary_size(), self.output_dim),
+            dtype=tf.float32,
+            initializer=self.embeddings_initializer,
+            name='embeddings',
+            regularizer=self.embeddings_regularizer,
+            constraint=self.embeddings_constraint,
+            experimental_autocast=False
+        )
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class NodeEmbeddingLookup(EmbeddingLookup):
     feature = 'node_feature'
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class EdgeEmbeddingLookup(EmbeddingLookup):
     feature = 'edge_feature'
