@@ -1,120 +1,45 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import layers
-from tensorflow.keras import initializers
-from keras.layers.preprocessing import preprocessing_utils as utils
+from keras import layers
+from keras import initializers
+
+import numpy as np
 
-from typing import Tuple
-from typing import List
 from typing import Optional
 from typing import Union
-from typing import Tuple
 
+from molgraph.internal import register_keras_serializable 
+from molgraph.internal import PreprocessingLayer
 
 from molgraph.tensors.graph_tensor import GraphTensor
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
-class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
+@register_keras_serializable(package='molgraph')
+class StandardScaling(PreprocessingLayer):
 
     '''Standard scaling, via centering and standardization.
 
     Specify, as keyword argument only,
     ``StandardScaling(feature='node_feature')`` to perform standard scaling
-    on the ``node_feature`` field of the ``GraphTensor``, or,
+    on the ``node_feature`` component of the ``GraphTensor``, or,
     ``StandardScaling(feature='edge_feature')`` to perform standard scaling
-    on the ``edge_feature`` field of the ``GraphTensor``. If not specified,
-    the ``node_feature`` field will be considered.
+    on the ``edge_feature`` component of the ``GraphTensor``. If not specified,
+    the ``node_feature`` component will be considered.
 
     **Examples:**
 
     Adapt layer on ``GraphTensor`` directly:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[2.0, 0.5], [2.0, 0.0]],
-    ...             [[2.0, 0.0], [2.0, 0.5], [0.0, 2.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[2., .5], [2., 0.], [2., 0.], [2., .5], [0., 2.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Initialize layer
     >>> preprocessing = molgraph.layers.StandardScaling(
     ...    feature='node_feature')
-    >>> # Adapt layer to graph_tensor
     >>> preprocessing.adapt(graph_tensor)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
-    >>> graph_tensor = model(graph_tensor)
-    >>> graph_tensor.merge().node_feature
-    <tf.Tensor: shape=(5, 2), dtype=float32, numpy=
-    array([[ 0.50000006, -0.1360828 ],
-           [ 0.50000006, -0.8164967 ],
-           [ 0.50000006, -0.8164967 ],
-           [ 0.50000006, -0.1360828 ],
-           [-2.0000002 ,  1.9051588 ]], dtype=float32)>
-
-    Adapt layer on ``tf.data.Dataset`` constructed from ``GraphTensor``:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[2.0, 0.5], [2.0, 0.0]],
-    ...             [[2.0, 0.0], [2.0, 0.5], [0.0, 2.0]]
-    ...         ],
-    ...     }
-    ... )
-    >>> # Obtain dataset
-    >>> ds = tf.data.Dataset.from_tensor_slices(graph_tensor).batch(2)
-    >>> # Initialize layer
-    >>> preprocessing = molgraph.layers.StandardScaling(
-    ...    feature='node_feature')
-    >>> # Adapt layer to graph_tensor
-    >>> preprocessing.adapt(ds)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
-    >>> output = model.predict(ds)
-    >>> output.merge().node_feature
-    <tf.Tensor: shape=(5, 2), dtype=float32, numpy=
-    array([[ 0.50000006, -0.1360828 ],
-           [ 0.50000006, -0.8164967 ],
-           [ 0.50000006, -0.8164967 ],
-           [ 0.50000006, -0.1360828 ],
-           [-2.0000002 ,  1.9051588 ]], dtype=float32)>
-
-    Adapt layer on merged ``GraphTensor``:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [2.0, 0.5],
-    ...             [2.0, 0.0],
-    ...             [2.0, 0.0],
-    ...             [2.0, 0.5],
-    ...             [0.0, 2.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Initialize layer
-    >>> preprocessing = molgraph.layers.StandardScaling(
-    ...    feature='node_feature')
-    >>> # Adapt layer to graph_tensor
-    >>> preprocessing.adapt(graph_tensor)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
+    >>> model = tf.keras.Sequential([preprocessing,])
     >>> graph_tensor = model(graph_tensor)
     >>> graph_tensor.node_feature
     <tf.Tensor: shape=(5, 2), dtype=float32, numpy=
@@ -124,6 +49,28 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
            [ 0.49999997, -0.1360828 ],
            [-2.        ,  1.9051588 ]], dtype=float32)>
 
+    Adapt layer on a tf.data.Dataset:
+
+    >>> graph_tensor = molgraph.GraphTensor(
+    ...     sizes=[2, 3],
+    ...     node_feature=[[2., .5], [2., 0.], [2., 0.], [2., .5], [0., 2.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
+    ... )
+    >>> ds = tf.data.Dataset.from_tensor_slices(graph_tensor).batch(2)
+    >>> preprocessing = molgraph.layers.StandardScaling(
+    ...    feature='node_feature')
+    >>> preprocessing.adapt(ds)
+    >>> model = tf.keras.Sequential([preprocessing])
+    >>> output = model.predict(ds, verbose=0)
+    >>> output.node_feature
+    <tf.Tensor: shape=(5, 2), dtype=float32, numpy=
+    array([[ 0.49999997, -0.1360828 ],
+           [ 0.49999997, -0.8164967 ],
+           [ 0.49999997, -0.8164967 ],
+           [ 0.49999997, -0.1360828 ],
+           [-2.        ,  1.9051588 ]], dtype=float32)>
+
     Args:
         mean (tf.Tensor, None):
             The means of the original data; used to transform the data.
@@ -178,13 +125,18 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
                 The number of steps of adaption. If None, the number of
                 samples divided by the batch_size is used. Default to None.
         '''
+
         if not isinstance(data,  GraphTensor):
-            data = data = data.map(
+            data = data.map(
                 lambda x: getattr(x, self.feature))
+            for x in data.take(1):
+                self.build(x.shape)
         else:
             data = getattr(data, self.feature)
-        super().adapt(data, batch_size=batch_size, steps=steps)
+            self.build(data.shape)
 
+        super().adapt(data, batch_size=batch_size, steps=steps)
+        
     def call(self, data):
         '''Defines the computation from inputs to outputs.
 
@@ -199,9 +151,10 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
         Returns:
             GraphTensor:
                 A ``GraphTensor`` with updated features. Either the
-                ``node_features`` field or the ``edge_features``
-                field (of the ``GraphTensor``) are updated.
+                ``node_feature`` component or the ``edge_feature``
+                component (of the ``GraphTensor``) are updated.
         '''
+
         feature = getattr(data, self.feature)
 
         if isinstance(feature, tf.RaggedTensor):
@@ -234,8 +187,8 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
         Args:
             input_shape (list, tuple, tf.TensorShape):
                 The shape of the input to the layer. Corresponds to either
-                the ``node_feature`` field or the ``edge_feature``
-                fields of ``GraphTensor``.
+                the ``node_feature`` component or the ``edge_feature``
+                component of ``GraphTensor``.
         '''
         super().build(input_shape)
 
@@ -262,17 +215,15 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
                 initializer=initializers.Constant(True),
                 trainable=False)
 
+        self.count = self.add_weight(
+            name='count',
+            shape=(),
+            dtype=tf.int64,
+            initializer='zeros',
+            trainable=False)
+            
         if self.input_mean is None:
-
-            self.count = self.add_weight(
-                name='count',
-                shape=(),
-                dtype=tf.int64,
-                initializer='zeros',
-                trainable=False)
-
             self.finalize_state()
-
         else:
             self.adapt_mean.assign(self.input_mean)
             self.adapt_variance.assign(self.input_variance)
@@ -292,7 +243,7 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
         Args:
             feature (tf.Tensor, tf.RaggedTensor):
                 A mini-batch of inputs to the layer. Corresponds to either
-                the ``node_feature`` or ``edge_feature`` field of
+                the ``node_feature`` or ``edge_feature`` component of
                 ``GraphTensor``.
         '''
 
@@ -384,15 +335,15 @@ class StandardScaling(layers.experimental.preprocessing.PreprocessingLayer):
 
     def get_config(self):
         config = {
-            'mean': utils.listify_tensors(self.mean),
-            'variance': utils.listify_tensors(self.variance),
+            'mean': _listify_tensors(self.mean),
+            'variance': _listify_tensors(self.variance),
             'feature': self.feature,
             'variance_threshold': self.variance_threshold}
         base_config = super().get_config()
         return dict(list(base_config.items()) + list(config.items()))
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class VarianceThreshold(StandardScaling):
 
     '''Variance thresholding.
@@ -401,101 +352,24 @@ class VarianceThreshold(StandardScaling):
     calling the layer.
 
     Specify, as keyword argument only,
-    ``VarianceThreshold(feature='node_feature')`` to perform standard scaling
-    on the ``node_feature`` field of the ``GraphTensor``, or,
-    ``VarianceThreshold(feature='edge_feature')`` to perform standard scaling
-    on the ``edge_feature`` field of the ``GraphTensor``. If not specified,
-    the ``node_feature`` field will be considered.
+    ``VarianceThreshold(feature='node_feature')`` to perform variance thresholding
+    on the ``node_feature`` component of the ``GraphTensor``, or,
+    ``VarianceThreshold(feature='edge_feature')`` to perform variance thresholding
+    on the ``edge_feature`` component of the ``GraphTensor``. If not specified,
+    the ``node_feature`` component will be considered.
 
-    **Examples:**
-
-    Adapt layer on ``GraphTensor`` directly:
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[2.0, 0.5], [2.0, 0.0]],
-    ...             [[2.0, 0.0], [2.0, 0.5], [0.0, 2.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[2., .5], [2., 0.], [2., 0.], [2., .5], [0., 2.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
-    >>> # Initialize layer
     >>> preprocessing = molgraph.layers.VarianceThreshold(
     ...    feature='node_feature', variance_threshold=0.6)
-    >>> # Adapt layer to graph_tensor
     >>> preprocessing.adapt(graph_tensor)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
-    >>> graph_tensor = model(graph_tensor)
-    >>> graph_tensor.merge().node_feature
-    <tf.Tensor: shape=(5, 1), dtype=float32, numpy=
-    array([[2.],
-           [2.],
-           [2.],
-           [2.],
-           [0.]], dtype=float32)>
-
-    Adapt layer on ``tf.data.Dataset`` constructed from ``GraphTensor``:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[2.0, 0.5], [2.0, 0.0]],
-    ...             [[2.0, 0.0], [2.0, 0.5], [0.0, 2.0]]
-    ...         ],
-    ...     }
-    ... )
-    >>> # Obtain dataset
-    >>> ds = tf.data.Dataset.from_tensor_slices(graph_tensor).batch(2)
-    >>> # Initialize layer
-    >>> preprocessing = molgraph.layers.VarianceThreshold(
-    ...    feature='node_feature', variance_threshold=0.6)
-    >>> # Adapt layer to graph_tensor
-    >>> preprocessing.adapt(ds)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
-    >>> output = model.predict(ds)
-    >>> output.merge().node_feature
-    <tf.Tensor: shape=(5, 1), dtype=float32, numpy=
-    array([[2.],
-           [2.],
-           [2.],
-           [2.],
-           [0.]], dtype=float32)>
-
-    Adapt layer on merged ``GraphTensor``:
-
-    >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [0, 1, 2, 2, 3, 3, 4, 4],
-    ...         'edge_src': [1, 0, 3, 4, 2, 4, 3, 2],
-    ...         'node_feature': [
-    ...             [2.0, 0.5],
-    ...             [2.0, 0.0],
-    ...             [2.0, 0.0],
-    ...             [2.0, 0.5],
-    ...             [0.0, 2.0]
-    ...         ],
-    ...         'graph_indicator': [0, 0, 1, 1, 1],
-    ...     }
-    ... )
-    >>> # Initialize layer
-    >>> preprocessing = molgraph.layers.VarianceThreshold(
-    ...    feature='node_feature', variance_threshold=0.6)
-    >>> # Adapt layer to graph_tensor
-    >>> preprocessing.adapt(graph_tensor)
-    >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
-    ...     preprocessing,
-    ... ])
+    >>> model = tf.keras.Sequential([preprocessing])
     >>> graph_tensor = model(graph_tensor)
     >>> graph_tensor.node_feature
     <tf.Tensor: shape=(5, 1), dtype=float32, numpy=
@@ -505,6 +379,28 @@ class VarianceThreshold(StandardScaling):
            [2.],
            [0.]], dtype=float32)>
 
+    Adapt layer on a tf.data.Dataset:
+
+    >>> graph_tensor = molgraph.GraphTensor(
+    ...     sizes=[2, 3],
+    ...     node_feature=[[2., .5], [2., 0.], [2., 0.], [2., .5], [0., 2.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
+    ... )
+    >>> ds = tf.data.Dataset.from_tensor_slices(graph_tensor).batch(2)
+    >>> preprocessing = molgraph.layers.VarianceThreshold(
+    ...    feature='node_feature', variance_threshold=0.6)
+    >>> preprocessing.adapt(ds)
+    >>> model = tf.keras.Sequential([preprocessing])
+    >>> output = model.predict(ds, verbose=0)
+    >>> output.node_feature
+    <tf.Tensor: shape=(5, 1), dtype=float32, numpy=
+    array([[2.],
+           [2.],
+           [2.],
+           [2.],
+           [0.]], dtype=float32)>
+
     Args:
         variance_threshold (int, float, None):
             The threshold for removing features, based on the variance of the
@@ -548,8 +444,8 @@ class VarianceThreshold(StandardScaling):
         Returns:
             GraphTensor:
                 A ``GraphTensor`` with updated features. Either the
-                ``node_features`` field or the ``edge_features``
-                field (of the ``GraphTensor``) are updated.
+                ``node_feature`` component or the ``edge_feature``
+                component (of the ``GraphTensor``) are updated.
         '''
         feature = getattr(data, self.feature)
 
@@ -564,21 +460,29 @@ class VarianceThreshold(StandardScaling):
         return data.update({self.feature: feature})
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class NodeStandardScaling(StandardScaling):
     feature = 'node_feature'
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class EdgeStandardScaling(StandardScaling):
     feature = 'edge_feature'
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class NodeVarianceThreshold(VarianceThreshold):
     feature = 'node_feature'
 
 
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class EdgeVarianceThreshold(VarianceThreshold):
     feature = 'edge_feature'
+
+
+def _listify_tensors(x):
+    if tf.is_tensor(x):
+        x = x.numpy()
+    if isinstance(x, np.ndarray):
+        x = x.tolist()
+    return x
