@@ -1,41 +1,35 @@
 import tensorflow as tf
 from tensorflow import keras
-from tensorflow.keras import layers
-from tensorflow.keras import activations
-from keras.utils import tf_utils
+from keras import layers
+from keras import activations
 
-from typing import Tuple
 from typing import Union
 from typing import Callable
 
+from molgraph.internal import register_keras_serializable 
+
 from molgraph.tensors.graph_tensor import GraphTensor
 
 
-
-@keras.utils.register_keras_serializable(package='molgraph')
+@register_keras_serializable(package='molgraph')
 class TransformerEncoderReadout(layers.Layer):
 
     '''Transformer encoder layer for graph readout.
 
-    **Example:**
+    Example usage:
 
     >>> graph_tensor = molgraph.GraphTensor(
-    ...     data={
-    ...         'edge_dst': [[0, 1], [0, 0, 1, 1, 2, 2]],
-    ...         'edge_src': [[1, 0], [1, 2, 0, 2, 1, 0]],
-    ...         'node_feature': [
-    ...             [[1.0, 0.0], [1.0, 0.0]],
-    ...             [[1.0, 0.0], [1.0, 0.0], [0.0, 1.0]]
-    ...         ],
-    ...     }
+    ...     sizes=[2, 3],
+    ...     node_feature=[[1., 0.], [1., 0.], [1., 0.], [1., 0.], [0., 1.]],
+    ...     edge_src=[1, 0, 3, 4, 2, 4, 3, 2],
+    ...     edge_dst=[0, 1, 2, 2, 3, 3, 4, 4],
     ... )
     >>> model = tf.keras.Sequential([
-    ...     tf.keras.layers.Input(type_spec=graph_tensor.unspecific_spec),
     ...     # molgraph.layers.GCNConv(4),
     ...     molgraph.layers.TransformerEncoderReadout()
     ... ])
-    >>> model.output_shape
-    (None, 2)
+    >>> model(graph_tensor).shape
+    TensorShape([2, 2])
 
     Args:
         hidden_units (int):
@@ -86,7 +80,7 @@ class TransformerEncoderReadout(layers.Layer):
 
         node_dim = self._node_feature_shape[-1]
 
-        with tf_utils.maybe_init_scope(self):
+        with tf.init_scope():
             self.attention = keras.layers.MultiHeadAttention(
                 num_heads=self.num_heads, key_dim=node_dim)
             self.attention._build_from_signature(
